import codecs
import datetime
import logging
import os
import re
import sys
from itertools import chain

import django
import yaml
from django.apps import apps
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from google.cloud.datastore.entity import Entity

from gcloudc.core.validators import MaxBytesValidator

from . import transaction
from .utils import get_top_concrete_parent

logger = logging.getLogger(__name__)
_project_special_indexes = {}
_app_special_indexes = {}
_last_loaded_times = {}
_indexes_loaded = False


MAX_COLUMNS_PER_SPECIAL_INDEX = getattr(settings, "DJANGAE_MAX_COLUMNS_PER_SPECIAL_INDEX", 3)
CHARACTERS_PER_COLUMN = [31, 44, 54, 63, 71, 79, 85, 91, 97, 103]
STRIP_PERCENTS = django.VERSION < (1, 10)


def _get_project_index_file(connection):
    project_index_file = os.path.join(connection.settings_dict["INDEXES_FILE"])
    return project_index_file


def _get_app_index_files():
    index_files = []

    for app_config in apps.get_app_configs():
        app_path = app_config.path
        project_index_file = os.path.join(app_path, "djangaeidx.yaml")
        index_files.append(project_index_file)
    return index_files


def _get_table_from_model(model_class):
    return model_class._meta.db_table


def _is_iterable(value):
    return hasattr(value, "__iter__") and not isinstance(value, (str, bytes))  # is a list, tuple or set?


def _deduplicate_list(value_list):
    """ Deduplicate list of elements; value_list is expected to be a list
    of containing hashable elements. """
    return list(set(value_list))


def _make_lower(value):
    """ Make string and list of strings lowercase """
    if _is_iterable(value):
        return [v.lower() for v in value]
    else:
        return value.lower()


def _merged_indexes():
    """
        Returns the combination of the app and project special indexes
    """
    global _project_special_indexes
    global _app_special_indexes

    result = _app_special_indexes.copy()
    for model, indexes in (_project_special_indexes or {}).items():
        for field_name, values in indexes.items():
            result.setdefault(model, {}).setdefault(field_name, []).extend(values)
    return result


def load_special_indexes(connection):
    global _project_special_indexes
    global _app_special_indexes
    global _last_loaded_times
    global _indexes_loaded

    RUNNING_DEVSERVER = len(sys.argv) > 1 and sys.argv[1] == "runserver"

    if _indexes_loaded and not RUNNING_DEVSERVER:
        # Index files can't change if we're on production, so once they're loaded we don't need
        # to check their modified times and reload them
        return

    def _read_file(filepath):
        # Load any existing indexes
        with open(filepath, "r") as stream:
            data = yaml.safe_load(stream)
        return data

    project_index_file = _get_project_index_file(connection)
    app_files = _get_app_index_files()

    files_to_reload = {}

    # Go through and reload any files that we find
    for file_path in [project_index_file] + app_files:
        if not os.path.exists(file_path):
            continue

        mtime = os.path.getmtime(file_path)
        if _last_loaded_times.get(file_path) and _last_loaded_times[file_path] == mtime:
            # The file hasn't changed since last time, so do nothing
            continue
        else:
            # Mark this file for reloading, store the current modified time
            files_to_reload[file_path] = mtime

    # First, reload the project index file,
    if project_index_file in files_to_reload:
        mtime = files_to_reload[project_index_file]
        _project_special_indexes = _read_file(project_index_file)
        _last_loaded_times[project_index_file] = mtime

        # Remove it from the files to reload
        del files_to_reload[project_index_file]

    # Now, load the rest of the files and update any entries
    for file_path in files_to_reload:
        mtime = files_to_reload[project_index_file]
        new_data = _read_file(file_path)
        _last_loaded_times[file_path] = mtime

        # Update the app special indexes list
        for model, indexes in new_data.items():
            for field_name, values in indexes.items():
                _app_special_indexes.setdefault(model, {}).setdefault(field_name, []).extend(values)

    _indexes_loaded = True
    logger.debug("Loaded special indexes for %d models", len(_merged_indexes()))


def special_index_exists(model_class, field_name, index_type):
    table = _get_table_from_model(model_class)
    return index_type in _merged_indexes().get(table, {}).get(field_name, [])


def special_indexes_for_model(model_class):
    classes = [model_class] + [x for x in model_class._meta.parents.keys()]

    result = {}
    for klass in classes:
        result.update(_merged_indexes().get(_get_table_from_model(klass), {}))
    return result


def special_indexes_for_column(model_class, column):
    return special_indexes_for_model(model_class).get(column, [])


def write_special_indexes(connection):
    """
        Writes the project-specific indexes to the project djangaeidx.yaml
    """
    project_index_file = _get_project_index_file(connection)

    with open(project_index_file, "w") as stream:
        stream.write(yaml.dump(_project_special_indexes))


def add_special_index(connection, model_class, field_name, indexer, operator, value=None):
    global _project_special_indexes

    index_type = indexer.prepare_index_type(operator, value)

    field_name = str(field_name)  # Make sure we are working with strings

    load_special_indexes(connection)

    if special_index_exists(model_class, field_name, index_type):
        return

    connection_params = connection.get_connection_params()

    # Test connections have a name starting with test_
    is_test_connection = connection_params["NAME"].startswith("test_")

    # Normal connections default to generating special indexes
    should_write_indexes = connection_params.get("GENERATE_SPECIAL_INDEXES", True)

    if is_test_connection:
        # Test connections default to NOT generating special indexes
        test_params = connection_params.get("TEST", {})
        should_write_indexes = test_params.get("GENERATE_SPECIAL_INDEXES", False)

    # FIXME: We need a can_write_indexes so we can throw an error if the filesystem isn't
    # writeable.
    if should_write_indexes:
        raise RuntimeError(
            "There is a missing index in your djangaeidx.yaml - \n\n{0}:\n\t{1}: [{2}]".format(
                _get_table_from_model(model_class), field_name, index_type
            )
        )

    if _project_special_indexes is None:
        _project_special_indexes = {}

    _project_special_indexes.setdefault(_get_table_from_model(model_class), {}).setdefault(field_name, []).append(
        str(index_type)
    )

    write_special_indexes(connection)


class IgnoreForIndexing(Exception):
    """
        An exception thrown from prep_value_for_database if the index column
        should be removed from the entity.

        Frustratingly, the legacy icontains and contains Indexers use 'value' when
        determining the column name, and so you must pass the prepared value for the
        column anyway, even though it won't be used.

        When the legacy icontains/contains indexers are removed we should:

         - Remove __init__ from this class
         - Remove 'value' as an argument to indexed_column_name
         - Stop setting 'values' from processed_value in django_instance_to_entity
    """

    def __init__(self, processed_value):
        self.processed_value = processed_value


class Indexer(object):
    # Set this to True if prep_value_for_database returns additional Entity instances
    # to save as descendents, rather than values to index as columns
    PREP_VALUE_RETURNS_ENTITIES = False

    # **IMPORTANT! If you return Entities from an indexer, the kind *must* start with
    # _djangae_idx_XXXX where XXX is the top concrete model kind of the instance you
    # are indexing. If you do not do this, then the tables will not be correctly flushed
    # when the database is flushed**

    @classmethod
    def cleanup(cls, client, datastore_key):
        """
            Called when an instance is deleted, if the instances has an index
            which uses this indexer. This is mainly for cleaning up descendent kinds
            (e.g. like those used in contains + icontains)
        """
        pass

    def handles(self, field, operator):
        """
            When given a field instance and an operator (e.g. gt, month__gt etc.)
            returns True or False whether or not this is the indexer to handle that
            situation
        """
        raise NotImplementedError()

    def validate_can_be_indexed(self, value, negated):
        """Return True if the value is indexable, False otherwise"""
        raise NotImplementedError()

    def prep_value_for_database(self, value, index, **kwargs):
        raise NotImplementedError()

    def prep_value_for_query(self, value, **kwargs):
        raise NotImplementedError()

    def indexed_column_name(self, field_column, value, index):
        raise NotImplementedError()

    def prep_query_operator(self, op):
        if "__" in op:
            return op.split("__")[-1]
        else:
            return "exact"  # By default do an exact operation

    def prepare_index_type(self, index_type, value):
        return index_type

    def unescape(self, value):
        value = value.replace("\\_", "_")
        value = value.replace("\\%", "%")
        value = value.replace("\\\\", "\\")
        return value


class StringIndexerMixin(object):
    STRING_FIELDS = (
        models.TextField,
        models.CharField,
        models.URLField,
        models.DateTimeField,  # Django allows these for some reason
        models.DateField,
        models.TimeField,
        models.IntegerField,  # SQL coerces ints to strings, and so we need these too
        models.PositiveIntegerField,
        models.AutoField,
    )

    def handles(self, field, operator):
        from gcloudc.db.models.fields import iterable

        try:
            # Make sure the operator is in there
            operator.split("__").index(self.OPERATOR)
        except ValueError:
            return False

        if isinstance(field, self.STRING_FIELDS):
            return True
        elif (
            isinstance(field, (iterable.ListField, iterable.SetField))
            and field.item_field_type.__class__ in self.STRING_FIELDS
            and operator.startswith("item__")
        ):
            return True
        return False


class DateIndexerMixin(object):
    def handles(self, field, operator):
        from gcloudc.db.models.fields import iterable

        DATE_FIELDS = (models.DateField, models.DateTimeField)

        if operator.split("__")[0] != self.OPERATOR:
            return False

        if isinstance(field, DATE_FIELDS):
            return True
        elif (
            isinstance(field, (iterable.ListField, iterable.SetField))
            and field.item_field_type.__class__ in DATE_FIELDS
            and operator.startswith("item__")
        ):
            return True

        return False


class TimeIndexerMixin(object):
    def handles(self, field, operator):
        from gcloudc.db.models.fields import iterable

        TIME_FIELDS = (models.TimeField, models.DateTimeField)

        if operator.split("__")[0] != self.OPERATOR:
            return False

        if isinstance(field, TIME_FIELDS):
            return True
        elif (
            isinstance(field, (iterable.ListField, iterable.SetField))
            and field.item_field_type.__class__ in TIME_FIELDS
            and operator.startswith("item__")
        ):
            return True

        return False


class IExactIndexer(StringIndexerMixin, Indexer):
    OPERATOR = "iexact"

    def validate_can_be_indexed(self, value, negated):
        return len(value) < 500

    def prep_value_for_database(self, value, index, **kwargs):
        if value is None:
            return None

        if isinstance(value, int):
            value = str(value)
        return value.lower()

    def prep_value_for_query(self, value, **kwargs):
        value = self.unescape(value)
        return value.lower()

    def indexed_column_name(self, field_column, value, index):
        return "_idx_iexact_{0}".format(field_column)


class HourIndexer(TimeIndexerMixin, Indexer):
    OPERATOR = "hour"

    def validate_can_be_indexed(self, value, negated):
        return isinstance(value, datetime.datetime)

    def prep_value_for_database(self, value, index, **kwargs):
        if value:
            return value.hour
        return None

    def prep_value_for_query(self, value, **kwargs):
        if isinstance(value, int):
            return value

        if isinstance(value, str):
            value = datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
        return value.hour

    def indexed_column_name(self, field_column, value, index):
        return "_idx_hour_{0}".format(field_column)


class MinuteIndexer(TimeIndexerMixin, Indexer):
    OPERATOR = "minute"

    def validate_can_be_indexed(self, value, negated):
        return isinstance(value, datetime.datetime)

    def prep_value_for_database(self, value, index, **kwargs):
        if value:
            return value.minute
        return None

    def prep_value_for_query(self, value, **kwargs):
        if isinstance(value, int):
            return value

        if isinstance(value, str):
            value = datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
        return value.minute

    def indexed_column_name(self, field_column, value, index):
        return "_idx_minute_{0}".format(field_column)


class SecondIndexer(TimeIndexerMixin, Indexer):
    OPERATOR = "second"

    def validate_can_be_indexed(self, value, negated):
        return isinstance(value, datetime.datetime)

    def prep_value_for_database(self, value, index, **kwargs):
        if value:
            return value.second
        return None

    def prep_value_for_query(self, value, **kwargs):
        if isinstance(value, int):
            return value

        if isinstance(value, str):
            value = datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
        return value.second

    def indexed_column_name(self, field_column, value, index):
        return "_idx_second_{0}".format(field_column)


class DayIndexer(DateIndexerMixin, Indexer):
    OPERATOR = "day"

    def validate_can_be_indexed(self, value, negated):
        return isinstance(value, (datetime.datetime, datetime.date))

    def prep_value_for_database(self, value, index, **kwargs):
        if value:
            return value.day
        return None

    def prep_value_for_query(self, value, **kwargs):
        if isinstance(value, int):
            return value

        if isinstance(value, str):
            value = datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
        return value.day

    def indexed_column_name(self, field_column, value, index):
        return "_idx_day_{0}".format(field_column)


class YearIndexer(DateIndexerMixin, Indexer):
    OPERATOR = "year"

    def validate_can_be_indexed(self, value, negated):
        return isinstance(value, (datetime.datetime, datetime.date))

    def prep_value_for_database(self, value, index, **kwargs):
        if value:
            return value.year
        return None

    def prep_value_for_query(self, value, **kwargs):
        if isinstance(value, int):
            return value

        if isinstance(value, str):
            value = datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")

        return value.year

    def indexed_column_name(self, field_column, value, index):
        return "_idx_year_{0}".format(field_column)


class MonthIndexer(DateIndexerMixin, Indexer):
    OPERATOR = "month"

    def validate_can_be_indexed(self, value, negated):
        return isinstance(value, (datetime.datetime, datetime.date))

    def prep_value_for_database(self, value, index, **kwargs):
        if value:
            return value.month
        return None

    def prep_value_for_query(self, value, **kwargs):
        if isinstance(value, int):
            return value

        if isinstance(value, str):
            value = datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")

        return value.month

    def indexed_column_name(self, field_column, value, index):
        return "_idx_month_{0}".format(field_column)


class WeekDayIndexer(DateIndexerMixin, Indexer):
    OPERATOR = "week_day"

    def validate_can_be_indexed(self, value, negated):
        return isinstance(value, (datetime.datetime, datetime.date))

    def prep_value_for_database(self, value, index, **kwargs):
        if value:
            zero_based_weekday = value.weekday()
            if zero_based_weekday == 6:  # Sunday
                return 1  # Django treats the week as starting at Sunday, but 1 based
            else:
                return zero_based_weekday + 2

        return None

    def prep_value_for_query(self, value, **kwargs):
        return value

    def indexed_column_name(self, field_column, value, index):
        return "_idx_week_day_{0}".format(field_column)


class ContainsIndexer(StringIndexerMixin, Indexer):
    PREP_VALUE_RETURNS_ENTITIES = True
    OPERATOR = u"contains"
    INDEXED_COLUMN_NAME = OPERATOR

    def validate_can_be_indexed(self, value, negated):
        if negated:
            return False

        try:
            MaxBytesValidator(limit_value=1500)(value)
            return True
        except ValidationError:
            return False

    @classmethod
    def cleanup(cls, client, datastore_key):

        # Kindless query, we don't know the kinds because we don't know all the fields
        # that use contains. But, we do know that all the things we need to delete are:
        # a.) A descendent
        # b.) Have a key name of whatever OPERATOR is

        qry = client.query(namespace=datastore_key.namespace, ancestor=datastore_key)
        qry.keys_only()

        # Delete all the entities matching the ancestor query
        client.delete([x.key for x in qry.fetch() if x.key.id_or_name == cls.OPERATOR])

    def _generate_kind_name(self, model, column):
        return "_djangae_idx_{}_{}".format(get_top_concrete_parent(model)._meta.db_table, column)

    def _generate_permutations(self, value):
        return [value[i:] for i in range(len(value))]

    def prep_value_for_database(self, value, index, model, column, connection):
        if value is None:
            raise IgnoreForIndexing([])

        # If this a date or a datetime, or something that supports isoformat, then use that
        if hasattr(value, "isoformat"):
            value = value.isoformat()

        if _is_iterable(value):
            value = list(chain(*[self._generate_permutations(v) for v in value]))
        else:
            value = self._generate_permutations(value)

        if not value:
            raise IgnoreForIndexing([])

        value = list(set(value))  # De-duplicate

        key = transaction._rpc(using=connection.alias).key(
            self._generate_kind_name(model, column), self.OPERATOR
        )
        entity = Entity(key)
        entity[self.INDEXED_COLUMN_NAME] = value
        return [entity]

    def prep_query_operator(self, operator):
        return "IN"

    def indexed_column_name(self, field_column, value, index):
        # prep_value_for_query returns a list PKs, so we return __key__ as the column
        return "__key__"

    def prep_value_for_query(self, value, model, column, connection):
        """
            Return a list of IDs of the associated contains models, these should
            match up with the IDs from the parent entities
        """

        if hasattr(value, "isoformat"):
            value = value.isoformat()
        else:
            value = str(value)
        value = self.unescape(value)

        if STRIP_PERCENTS:
            # SQL does __contains by doing LIKE %value%
            if value.startswith("%") and value.endswith("%"):
                value = value[1:-1]

        namespace = connection.namespace

        qry = transaction._rpc(using=connection.alias).query(
            kind=self._generate_kind_name(model, column), namespace=namespace
        )
        qry.keys_only()

        qry.add_filter(self.INDEXED_COLUMN_NAME, ">=", value)
        qry.add_filter(self.INDEXED_COLUMN_NAME, "<=", value + u"\ufffd")

        # We can't filter on the 'name' as part of the query, because the name is the key and these
        # are child entities of the ancestor entities which they are indexing, and as we don't know
        # the keys of the ancestor entities we can't create the complete keys, hence the comparison
        # of `x.name() == self.OPERATOR` happens here in python
        resulting_keys = set([x.key.parent for x in qry.fetch() if x.key.name == self.OPERATOR])
        return resulting_keys


class IContainsIndexer(ContainsIndexer):
    OPERATOR = "icontains"

    def _generate_permutations(self, value):
        return super(IContainsIndexer, self)._generate_permutations(value.lower())

    def prep_value_for_query(self, value, model, column, connection):
        return super(IContainsIndexer, self).prep_value_for_query(value.lower(), model, column, connection)


class LegacyContainsIndexer(StringIndexerMixin, Indexer):
    OPERATOR = "contains"

    def number_of_permutations(self, value):
        return sum(range(len(value) + 1))

    def validate_can_be_indexed(self, value, negated):
        if negated:
            return False
        return isinstance(value, str) and len(value) <= 500

    def prep_value_for_database(self, value, index, **kwargs):
        results = []
        if value:
            # If this a date or a datetime, or something that supports isoformat, then use that
            if hasattr(value, "isoformat"):
                value = value.isoformat()

            if self.number_of_permutations(value) > MAX_COLUMNS_PER_SPECIAL_INDEX * 500:
                raise ValueError(
                    "Can't index for contains query, this value is too long and has too many "
                    "permutations. You can increase the DJANGAE_MAX_COLUMNS_PER_SPECIAL_INDEX "
                    "setting to fix that. Use with caution."
                )
            if len(value) > CHARACTERS_PER_COLUMN[-1]:
                raise ValueError(
                    "Can't index for contains query, this value can be maximum {0} characters long.".format(
                        CHARACTERS_PER_COLUMN[-1]
                    )
                )

            if _is_iterable(value):
                # `value` is a list of strings. Generate a single combined list containing the
                # substrings of each string in `value`
                for element in value:
                    length = len(element)
                    lists = [element[i: j + 1] for i in range(length) for j in range(i, length)]
                    results.extend(lists)
            else:
                # `value` is a string. Generate a list of all its substrings.
                length = len(value)
                lists = [value[i: j + 1] for i in range(length) for j in range(i, length)]
                results.extend(lists)

        if not results:
            raise IgnoreForIndexing([])

        return _deduplicate_list(results)

    def prep_value_for_query(self, value, **kwargs):
        if hasattr(value, "isoformat"):
            value = value.isoformat()
        else:
            value = str(value)
        value = self.unescape(value)

        if STRIP_PERCENTS:
            # SQL does __contains by doing LIKE %value%
            if value.startswith("%") and value.endswith("%"):
                value = value[1:-1]

        return value

    def indexed_column_name(self, field_column, value, index):
        # This we use when we actually query to return the right field for a given
        # value length
        length = len(value)

        column_number = 0
        for x in CHARACTERS_PER_COLUMN:
            if length > x:
                column_number += 1
        return "_idx_contains_{0}_{1}".format(field_column, column_number)

    def prep_query_operator(self, op):
        return "exact"


class LegacyIContainsIndexer(LegacyContainsIndexer):
    OPERATOR = "icontains"

    def prep_value_for_database(self, value, index, **kwargs):
        if value is None:
            raise IgnoreForIndexing("")
        value = _make_lower(value)
        result = super(LegacyIContainsIndexer, self).prep_value_for_database(value, index)
        return result if result else None

    def indexed_column_name(self, field_column, value, index):
        column_name = super(LegacyIContainsIndexer, self).indexed_column_name(field_column, value, index)
        return column_name.replace("_idx_contains_", "_idx_icontains_")

    def prep_value_for_query(self, value, **kwargs):
        return super(LegacyIContainsIndexer, self).prep_value_for_query(value).lower()


class EndsWithIndexer(StringIndexerMixin, Indexer):
    """
        dbindexer originally reversed the string and did a startswith on it.
        However, this is problematic as it uses an inequality and therefore
        limits the queries you can perform. Instead, we store all permutations
        of the last characters in a list field. Then we can just do an exact lookup on
        the value. Which isn't as nice, but is more flexible.
    """

    OPERATOR = "endswith"

    def validate_can_be_indexed(self, value, negated):
        if negated:
            return False

        return isinstance(value, str) and len(value) < 500

    def prep_value_for_database(self, value, index, **kwargs):
        if value is None:
            return None

        results = []
        if _is_iterable(value):
            # `value` is a list of strings. Create a single combined list of "endswith" values
            # of all the strings in the list
            for element in value:
                for i in range(0, len(element)):
                    results.append(element[i:])
        else:
            # `value` is a string. Create a list of "endswith" strings.
            for i in range(0, len(value)):
                results.append(value[i:])

        if not results:
            return None

        return _deduplicate_list(results)

    def prep_value_for_query(self, value, **kwargs):
        value = self.unescape(value)
        if STRIP_PERCENTS:
            if value.startswith("%"):
                value = value[1:]
        return value

    def indexed_column_name(self, field_column, value, index):
        return "_idx_endswith_{0}".format(field_column)

    def prep_query_operator(self, op):
        return "exact"


class IEndsWithIndexer(EndsWithIndexer):
    """
        Same as above, just all lower cased
    """

    OPERATOR = "iendswith"

    def prep_value_for_database(self, value, index, **kwargs):
        if value:
            value = _make_lower(value)
        return super(IEndsWithIndexer, self).prep_value_for_database(value, index)

    def prep_value_for_query(self, value, **kwargs):
        return super(IEndsWithIndexer, self).prep_value_for_query(value.lower())

    def indexed_column_name(self, field_column, value, index):
        return "_idx_iendswith_{0}".format(field_column)


class StartsWithIndexer(StringIndexerMixin, Indexer):
    """
        Although we can do a startswith natively, doing it this way allows us to
        use more queries (E.g. we save an exclude)
    """

    OPERATOR = "startswith"

    def validate_can_be_indexed(self, value, negated):
        if negated:
            return False

        return isinstance(value, str) and len(value) < 500

    def prep_value_for_database(self, value, index, **kwargs):
        if value is None:
            return None

        if isinstance(value, datetime.datetime):
            value = value.strftime("%Y-%m-%d %H:%M:%S")

        results = []
        if _is_iterable(value):
            # `value` is a list of strings. Create a single combined list of "startswith" values
            # of all the strings in the list
            for element in value:
                for i in range(1, len(element) + 1):
                    results.append(element[:i])
        else:
            # `value` is a string. Create a list of "startswith" strings.
            for i in range(1, len(value) + 1):
                results.append(value[:i])

        if not results:
            return None

        return _deduplicate_list(results)

    def prep_value_for_query(self, value, **kwargs):
        value = self.unescape(value)
        if STRIP_PERCENTS:
            if value.endswith("%"):
                value = value[:-1]
        return value

    def indexed_column_name(self, field_column, value, index):
        return "_idx_startswith_{0}".format(field_column)

    def prep_query_operator(self, op):
        return "exact"


class IStartsWithIndexer(StartsWithIndexer):
    """
        Same as above, just all lower cased
    """

    OPERATOR = "istartswith"

    def prep_value_for_database(self, value, index, **kwargs):
        if value:
            value = _make_lower(value)
        return super(IStartsWithIndexer, self).prep_value_for_database(value, index)

    def prep_value_for_query(self, value, **kwargs):
        return super(IStartsWithIndexer, self).prep_value_for_query(value.lower())

    def indexed_column_name(self, field_column, value, index):
        return "_idx_istartswith_{0}".format(field_column)


def _to_hex(value):
    value = codecs.encode(
        value.encode("utf8"), "hex"
    ).decode("utf8")
    return value


def _from_hex(value):
    return codecs.decode(
        value.encode("utf8"), "hex"
    ).decode("utf8")


class RegexIndexer(StringIndexerMixin, Indexer):
    OPERATOR = "regex"

    def prepare_index_type(self, index_type, value):
        """
            If we're dealing with RegexIndexer, we create a new index for each
            regex pattern. Indexes are called regex__pattern.
        """
        return "{}__{}".format(index_type, _to_hex(value))

    def validate_can_be_indexed(self, value, negated):
        if negated:
            return False

        return isinstance(value, bool)

    def get_pattern(self, index):
        try:
            return _from_hex(index.split("__")[-1])
        except IndexError:
            return ""

    def check_if_match(self, value, index, flags=0):
        pattern = self.get_pattern(index)

        if value:
            if _is_iterable(value):
                if any([bool(re.search(pattern, x, flags)) for x in value]):
                    return True
            else:
                if isinstance(value, int):
                    value = str(value)

                return bool(re.search(pattern, value, flags))
        return False

    def prep_value_for_database(self, value, index, **kwargs):
        return self.check_if_match(value, index)

    def prep_value_for_query(self, value, **kwargs):
        return True

    def indexed_column_name(self, field_column, value, index):
        return "_idx_regex_{0}_{1}".format(field_column, _to_hex(self.get_pattern(index)))

    def prep_query_operator(self, op):
        return "exact"


class IRegexIndexer(RegexIndexer):
    OPERATOR = "iregex"

    def prepare_index_type(self, index_type, value):
        return "{}__{}".format(index_type, _to_hex(value))

    def prep_value_for_database(self, value, index, **kwargs):
        return self.check_if_match(value, index, flags=re.IGNORECASE)

    def indexed_column_name(self, field_column, value, index):
        return "_idx_iregex_{0}_{1}".format(field_column, _to_hex(self.get_pattern(index)))


_REGISTERED_INDEXERS = []


def register_indexer(indexer_class):
    global _REGISTERED_INDEXERS
    _REGISTERED_INDEXERS.append(indexer_class())


def get_indexer(field, operator):
    global _REGISTERED_INDEXERS

    for indexer in _REGISTERED_INDEXERS:
        if indexer.handles(field, operator):
            return indexer


def indexers_for_model(model_class):
    indexes = special_indexes_for_model(model_class)

    indexers = []
    for field in model_class._meta.fields:
        if field.name in indexes:
            for operator in indexes[field.name]:
                indexers.append(get_indexer(field, operator))
    return set(indexers)


register_indexer(IExactIndexer)

if getattr(settings, "DJANGAE_USE_LEGACY_CONTAINS_LOGIC", False):
    register_indexer(LegacyContainsIndexer)
    register_indexer(LegacyIContainsIndexer)
else:
    register_indexer(ContainsIndexer)
    register_indexer(IContainsIndexer)

register_indexer(HourIndexer)
register_indexer(MinuteIndexer)
register_indexer(SecondIndexer)
register_indexer(DayIndexer)
register_indexer(MonthIndexer)
register_indexer(YearIndexer)
register_indexer(WeekDayIndexer)
register_indexer(EndsWithIndexer)
register_indexer(IEndsWithIndexer)
register_indexer(StartsWithIndexer)
register_indexer(IStartsWithIndexer)
register_indexer(RegexIndexer)
register_indexer(IRegexIndexer)



"""
JSONField automatically serializes most Python terms to JSON data.
Creates a TEXT field with a default value of "{}".  See test_json.py for
more information.

 from django.db import models
 from django_extensions.db.fields import json

 class LOL(models.Model):
     extra = json.JSONField()

This field originated from the django_extensions project: https://github.com/django-extensions/django-extensions
"""

from __future__ import absolute_import

import json
from collections import OrderedDict

from django.db import models
from django.core.serializers.json import DjangoJSONEncoder

from gcloudc.db.backends.datastore.indexing import Indexer, register_indexer, IgnoreForIndexing
from gcloudc.forms.fields import JSONFormField, JSONWidget

__all__ = ("JSONField",)


def dumps(value):
    return DjangoJSONEncoder().encode(value)


def loads(txt, object_pairs_hook=None):
    value = json.loads(txt, object_pairs_hook=object_pairs_hook)
    return value


class JSONDict(dict):
    """
    Hack so repr() called by dumpdata will output JSON instead of
    Python formatted data.  This way fixtures will work!
    """

    def __repr__(self):
        return dumps(self)


class JSONUnicode(str):
    """
    As above
    """

    def __repr__(self):
        return dumps(self)


class JSONList(list):
    """
    As above
    """

    def __repr__(self):
        return dumps(self)


class JSONOrderedDict(OrderedDict):
    """
    As above
    """

    def __repr__(self):
        return dumps(self)


class JSONKeyLookup(models.Lookup):
    lookup_name = "json_path"
    operator = "json_path"
    bilateral_transforms = False
    lookup_supports_text = True  # Tell Djangae connector that we are OK on text fields

    def __init__(self, path):
        self.path = path

    def get_rhs_op(self, connection, rhs):
        return "__".join([self.operator, self.path])

    def __call__(self, lhs, rhs):
        self.lhs = lhs
        self.rhs = rhs
        return self


class JSONField(models.TextField):
    """JSONField is a generic textfield that neatly serializes/unserializes
    JSON objects seamlessly.  Main thingy must be a dict object."""

    def __init__(self, use_ordered_dict=False, *args, **kwargs):
        if "default" in kwargs:
            if not callable(kwargs["default"]):
                raise TypeError("'default' must be a callable (e.g. 'dict' or 'list')")
        else:
            kwargs["default"] = dict

        # use `collections.OrderedDict` rather than built-in `dict`
        self.use_ordered_dict = use_ordered_dict

        models.TextField.__init__(self, *args, **kwargs)

    def parse_json(self, value):
        """Convert our string value to JSON after we load it from the DB"""
        if value is None or value == "":
            return {}
        elif isinstance(value, str):
            try:
                if self.use_ordered_dict:
                    res = loads(value, object_pairs_hook=OrderedDict)
                else:
                    res = loads(value)
            except ValueError:
                # If we can't parse as JSON, just return the value as-is
                return value

            if isinstance(res, OrderedDict) and self.use_ordered_dict:
                return JSONOrderedDict(res)
            elif isinstance(res, dict):
                return JSONDict(**res)
            elif isinstance(res, str):
                return JSONUnicode(res)
            elif isinstance(res, list):
                return JSONList(res)
            return res
        else:
            return value

    def to_python(self, value):
        return self.parse_json(value)

    def from_db_value(self, value, expression, connection, **kwargs):
        return self.parse_json(value)

    def value_to_string(self, obj):
        value = self.value_from_object(obj)

        if isinstance(value, str):
            return value

        return dumps(value)

    def get_db_prep_save(self, value, connection, **kwargs):
        """Convert our JSON object to a string before we save"""
        if value is None and self.null:
            return None

        return super(JSONField, self).get_db_prep_save(dumps(value), connection=connection)

    def south_field_triple(self):
        """Returns a suitable description of this field for South."""
        # We'll just introspect the _actual_ field.
        from south.modelsinspector import introspector

        field_class = "django.db.models.fields.TextField"
        args, kwargs = introspector(self)
        # That's our definition!
        return (field_class, args, kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(JSONField, self).deconstruct()
        if self.default == {}:
            del kwargs["default"]
        return name, path, args, kwargs

    def formfield(self, **kwargs):
        defaults = {"form_class": JSONFormField, "widget": JSONWidget}
        defaults.update(kwargs)
        return super(JSONField, self).formfield(**defaults)

    def get_lookup(self, lookup):
        ret = super(JSONField, self).get_lookup(lookup)
        if ret:
            return ret
        return JSONKeyLookup(lookup)

    def get_transform(self, lookup):
        ret = super(JSONField, self).get_transform(lookup)
        if ret:
            return ret

        # Assume a key path lookup
        class LookupBuilder(models.Transform):
            source_expressions = []
            lookup_name = "json_path"

            def __init__(self, *args, **kwargs):
                super(LookupBuilder, self).__init__(*args, **kwargs)
                self.path = [lookup]

            def get_transform(self, lookup):
                self.path.append(lookup)
                return lambda lhs, *args, **kwargs: self

            def get_lookup(self, lookup):
                self.path.append(lookup)
                return JSONKeyLookup("__".join(self.path))

        return LookupBuilder


class JSONKeyLookupIndexer(Indexer):
    OPERATOR = "json_path"

    def handles(self, field, operator):
        from gcloudc.db.models.fields.json import JSONField

        operator_part = operator.split("__", 1)[0]
        return isinstance(field, JSONField) and operator_part == self.OPERATOR

    def prepare_index_type(self, index_type, value):
        return index_type

    def prep_value_for_query(self, value, **kwargs):
        return value

    def prep_query_operator(self, op):
        return "exact"

    def prep_value_for_database(self, value, index, **kwargs):
        if isinstance(value, str):
            value = json.loads(value)

        index_part = index.split("__", 1)[1]
        path = index_part.split("__")

        is_isnull = False
        # Ignore isnull on the end of a path, it's not a value lookup
        if len(path) > 1 and path[-1] == "isnull":
            is_isnull = True
            path.pop()

        # Go through the path and look up the value from each
        # dictionary/list if we fail to find a value we raise
        # a IgnoreForIndexing exception which tells the special indexer
        # to not save *anything*
        for i, section in enumerate(path):
            try:
                section = int(section)
            except (TypeError, ValueError):
                pass

            try:
                value = value[section]
            except (KeyError, IndexError, TypeError):
                raise IgnoreForIndexing("")

        if is_isnull:
            return value is None

        return value

    def indexed_column_name(self, field_column, value, index):
        return "_idx_json_path_{}_{}".format(field_column, index.split("__", 1)[-1])


# Especially for JSON fields
register_indexer(JSONKeyLookupIndexer)



"""
    Utility functions for gathering data from Google Datastore Query
    objects.
"""


def has_filter(query, col_and_operator):
    """
        query: A Cloud Datastore Query object
        col_and_operator: tuple of column name and operator
    """
    for col, operator, value in query.filters:
        if (col, operator) == tuple(col_and_operator):
            return True

    return False


def get_filter(query, col_and_operator):
    """
        query: A Cloud Datastore Query object
        col_and_operator: tuple of column name and operator
    """
    for col, operator, value in query.filters:
        if (col, operator) == tuple(col_and_operator):
            return value

    return None


def is_keys_only(query):
    return query.projection == ["__key__"]


def compare_keys(lhs, rhs):
    """
        The App Engine API used to provide a key comparison, but for
        some reason the Cloud Datastore API doesn't :(
    """

    def cmp(a, b):
        if a is None and b is None:
            return 0

        if a is None:
            return -1

        if b is None:
            return 1

        return (a > b) - (a < b)

    lhs_args = [lhs.project, lhs.namespace] + list(lhs.flat_path)
    if lhs.is_partial:
        # If the key is partial, then we need to add a blank placeholder
        # for the id or name so we can compare correctly
        lhs_args.extend("")

    rhs_args = [rhs.project, rhs.namespace] + list(rhs.flat_path)
    if rhs.is_partial:
        rhs_args.extend("")

    for lhs_component, rhs_component in zip(lhs_args, rhs_args):
        comparison = cmp(lhs_component, rhs_component)
        if comparison != 0:
            return comparison

    return cmp(len(lhs_args), len(rhs_args))



from django.db import models

from . import TestCase


class ExcludedPKModel(models.Model):
    name = models.CharField(max_length=30, primary_key=True)
    color = models.CharField(max_length=30)


class ExcludingPKTests(TestCase):
    def test_exclude_pks_with_slice(self):
        for i in range(10):
            ExcludedPKModel.objects.create(name=str(i), color=str(i))

        to_exclude = [str(x) for x in list(range(5)) + list(range(15, 20))]

        to_return = ExcludedPKModel.objects.exclude(pk__in=set(to_exclude)).values_list("pk", flat=True)[:2]
        self.assertEqual(2, len(to_return))

        qs = ExcludedPKModel.objects.filter(pk__in=to_return)

        self.assertEqual(2, len(qs))

    def test_count_on_excluded_pks(self):
        ExcludedPKModel.objects.create(name="Apple", color="Red")
        ExcludedPKModel.objects.create(name="Orange", color="Orange")

        self.assertEqual(
            1, ExcludedPKModel.objects.filter(pk__in=["Apple", "Orange"]).exclude(pk__in=["Apple"]).count()
        )



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

NAME = 'django-gcloud-connectors'

DESCRIPTION = 'A Django library for connecting to Google Cloud Datastore from Python 3 runtimes.'
URL = 'https://gitlab.com/potato-oss/google-cloud/django-gcloud-connectors'
LONG_DESCRIPTION = open(os.path.join(os.path.dirname(__file__), 'README.md')).read()

AUTHOR = "Potato London Ltd."
AUTHOR_EMAIL = "mail@p.ota.to"

if os.environ.get('CI_COMMIT_TAG'):
    VERSION = os.environ['CI_COMMIT_TAG']
else:
    VERSION = '0.3.6'

setup(
    name=NAME,
    version=VERSION,
    packages=find_packages(),
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    keywords=["Google Cloud Datastore", "Google App Engine", "Django"],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    install_requires=[
        'django>=2.0,<=3.3.*',
        'pyyaml>=5.3.1,<6.0',
        'google-cloud-datastore>=1.15.3',
        'sleuth-mock==0.1',
        'pyuca==1.2',
    ],
    include_package_data=True,
)



from django.core.exceptions import NON_FIELD_ERRORS

from .dbapi import NotSupportedError


MAX_ALLOWABLE_QUERIES = 30


class UniquenessMixin(object):
    """
    Mixin overriding the default methods checking value uniqueness (the logic
    if a copy of the Django implementation, aside from the part marked by
    comments).

    This must be used if you have a list or set field marked as unique.
    """

    def _perform_unique_checks(self, unique_checks):
        errors = {}
        for model_class, unique_check in unique_checks:
            lookup_kwargs = {}
            for field_name in unique_check:
                f = self._meta.get_field(field_name)
                lookup_value = getattr(self, f.attname)
                if lookup_value is None:
                    continue
                if f.primary_key and not self._state.adding:
                    continue

                ##########################################################################
                # This is a modification to Django's native implementation of this method;
                # we conditionally build a __in lookup if the value is an iterable.
                lookup = str(field_name)
                if isinstance(lookup_value, (list, set, tuple)):
                    lookup = "%s__overlap" % lookup

                lookup_kwargs[lookup] = lookup_value
                ##########################################################################
                # / end of changes

            if len(unique_check) != len(lookup_kwargs):
                continue

            #######################################################
            # Deal with long __in lookups by doing multiple queries in that case
            # This is a bit hacky, but we really have no choice due to App Engine's
            # 30 multi-query limit. This also means we can't support multiple list fields in
            # a unique combination
            #######################################################

            if len([x for x in lookup_kwargs if x.endswith("__in")]) > 1:
                raise NotSupportedError("You cannot currently have two list fields in a unique combination")

            # Split IN queries into multiple lookups if they are too long
            lookups = []
            for k, v in lookup_kwargs.items():
                if (k.endswith("__in") or k.endswith("__overlap")) and len(v) > MAX_ALLOWABLE_QUERIES:
                    v = list(v)
                    while v:
                        new_lookup = lookup_kwargs.copy()
                        new_lookup[k] = v[:30]
                        v = v[30:]
                        lookups.append(new_lookup)
                    break
            else:
                # Otherwise just use the one lookup
                lookups = [lookup_kwargs]

            for lookup_kwargs in lookups:
                qs = model_class._default_manager.filter(**lookup_kwargs).values_list("pk", flat=True)
                model_class_pk = self._get_pk_val(model_class._meta)
                result = list(qs)

                if not self._state.adding and model_class_pk is not None:
                    # If we are saving an instance, we ignore it's PK in the result
                    try:
                        result.remove(model_class_pk)
                    except ValueError:
                        pass

                if result:
                    if len(unique_check) == 1:
                        key = unique_check[0]
                    else:
                        key = NON_FIELD_ERRORS
                    errors.setdefault(key, []).append(self.unique_error_message(model_class, unique_check))
                    break
        return errors



from . import TestCase
from .models import TestUser


class DjangoInteractionTests(TestCase):

    def test_update_or_create_works(self):
        """
            update_or_create uses Django's atomic()
        """

        user, created = TestUser.objects.update_or_create(
            username="test",
            defaults={
                "email": "test@example.com"
            }
        )

        user_id = user.pk

        self.assertTrue(created)
        self.assertTrue(user_id)

        user, created = TestUser.objects.update_or_create(
            username="test",
            defaults={
                "email": "test2@example.com"
            }
        )

        self.assertFalse(created)
        self.assertEqual(user.pk, user_id)



# encoding: utf-8
import logging
import os
import zipfile

from django.db import models

from .charfields import CharField


# pyuca only supports version 5.2.0 of the collation algorithm on Python 2.x
COLLATION_FILE = "allkeys-5.2.0.txt"
COLLATION_ZIP_FILE = os.path.join(os.path.dirname(__file__), "allkeys-5.2.0.zip")

logger = logging.getLogger(__file__)


class ZipLoaderMixin(object):
    """
        The UCA collation file is massive (nearly 1.5M) but it's all text
        so it compresses easily. We ship the file zipped up and then decompress
        it on the fly here to save on storage, data transfer, memory etc.
        The use of generators on load should be efficient.
    """

    def __init__(self, zip_filename, text_filename):
        """
            The BaseCollator class __init__ takes a filename and calls
            load(filename). Here we pass up the text filename but store the
            zip filename, then override load so we can load from the zip instead
            of a filesystem.
        """
        self.zip_filename = zip_filename
        super(ZipLoaderMixin, self).__init__(filename=text_filename)

    def load(self, filename):
        from pyuca.collator import COLL_ELEMENT_PATTERN, hexstrings2int  # pyuca is required for ComputedCollationField

        with zipfile.ZipFile(self.zip_filename) as z:
            with z.open(filename) as f:
                for line in f.readlines():
                    line = line.decode("utf-8")
                    line = line.split("#", 1)[0].rstrip()

                    if not line or line.startswith("@version"):
                        continue

                    a, b = line.split(";", 1)
                    char_list = hexstrings2int(a.split())
                    coll_elements = []
                    for x in COLL_ELEMENT_PATTERN.finditer(b.strip()):
                        weights = x.groups()
                        coll_elements.append(hexstrings2int(weights))
                    self.table.add(char_list, coll_elements)


class ComputedFieldMixin:
    def __init__(self, func, *args, **kwargs):
        self.computer = func

        kwargs["editable"] = False

        super().__init__(*args, **kwargs)

    def pre_save(self, model_instance, add):
        value = self.get_computed_value(model_instance)
        setattr(model_instance, self.attname, value)
        return value

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        args = [self.computer] + args
        del kwargs["editable"]
        return name, path, args, kwargs

    def from_db_value(self, value, expression, connection, **kwargs):
        if value is None:
            return value
        return self.to_python(value)

    def get_computed_value(self, model_instance):
        # self.computer is either a function or a string containing the name of a method
        if callable(self.computer):
            return self.computer(model_instance)
        else:
            computer = getattr(model_instance, self.computer)
            return computer()


class ComputedCharField(ComputedFieldMixin, models.CharField):
    pass


class ComputedIntegerField(ComputedFieldMixin, models.IntegerField):
    pass


class ComputedTextField(ComputedFieldMixin, models.TextField):
    pass


class ComputedPositiveIntegerField(ComputedFieldMixin, models.PositiveIntegerField):
    pass


class ComputedBooleanField(ComputedFieldMixin, models.BooleanField):
    pass


class ComputedNullBooleanField(ComputedFieldMixin, models.NullBooleanField):
    pass


class ComputedCollationField(ComputedFieldMixin, CharField):
    """
        App Engine sorts strings based on the unicode codepoints that make them
        up. When you have strings from non-ASCII languages this makes the sort order
        incorrect (e.g. Ł will be sorted after Z).
        This field uses the pyuca library to calculate a sort key using the
        Unicode Collation Algorithm, which can then be used for ordering querysets
        correctly.
    """

    collator = None

    def __init__(self, source_field_name):
        import pyuca  # noqa: F401 Required dependency for ComputedCollationField
        from pyuca.collator import Collator_5_2_0

        # Instantiate Collator once only to save on memory / processing
        if not ComputedCollationField.collator:

            class Collator(ZipLoaderMixin, Collator_5_2_0):
                pass

            ComputedCollationField.collator = Collator(COLLATION_ZIP_FILE, COLLATION_FILE)

        def truncate(unicode_str):
            encoded = unicode_str.encode("utf-8")[:1500]
            # We ignore unrecognized chars as the truncation might
            # have split a unicode char down the middle
            return encoded.decode("utf-8", "ignore")

        def computer(instance):
            source_value = getattr(instance, source_field_name) or u""
            if not isinstance(source_value, str):
                source_value = str(source_value, "utf-8")
            sort_key = self.collator.sort_key(source_value)
            sort_key = u"".join([chr(x) for x in sort_key])
            truncated_key = truncate(sort_key)
            if truncated_key != sort_key:
                logger.warn("Truncated sort key for '%s.%s'", instance._meta.db_table, source_field_name)
            return truncated_key

        super(ComputedCollationField, self).__init__(computer)

    def deconstruct(self):
        name, path, args, kwargs = super(ComputedCollationField, self).deconstruct()
        del kwargs["max_length"]
        return name, path, args, kwargs



from django.core.validators import (
    BaseValidator,
    MaxLengthValidator,
    MinLengthValidator,
)
from django.utils.deconstruct import deconstructible
from django.utils.translation import ungettext_lazy


class MaxBytesValidator(BaseValidator):
    def compare(self, a, b):
        return a > b

    def clean(self, x):
        return len(x.encode("utf-8"))

    message = ungettext_lazy(
        "Ensure this value has at most %(limit_value)d byte (it has %(show_value)d).",
        "Ensure this value has at most %(limit_value)d bytes (it has %(show_value)d).",
        "limit_value",
    )
    code = "max_length"


@deconstructible
class MinItemsValidator(MinLengthValidator):
    """ Copy of MinLengthValidator, but with the message customised to say "items" instead of
        "characters".
    """

    message = ungettext_lazy(
        "Ensure this field has at least %(limit_value)d item (it has %(show_value)d).",
        "Ensure this field has at least %(limit_value)d items (it has %(show_value)d).",
        "limit_value",
    )


@deconstructible
class MaxItemsValidator(MaxLengthValidator):
    """ Copy of MaxLengthValidator, but with the message customised to say "items" instead of
        "characters".
    """

    message = ungettext_lazy(
        "Ensure this field has at most %(limit_value)d item (it has %(show_value)d).",
        "Ensure this field has at most %(limit_value)d items (it has %(show_value)d).",
        "limit_value",
    )



import sleuth
from django.db import NotSupportedError
from django.db.models import Q
from django.test import override_settings

from . import TestCase
from .models import (
    MultiQueryModel,
    NullableFieldModel,
)


class AsyncMultiQueryTest(TestCase):
    """
        Specific tests for multiquery
    """

    def test_hundred_or(self):
        for i in range(100):
            MultiQueryModel.objects.create(field1=i)

        self.assertEqual(len(MultiQueryModel.objects.filter(field1__in=list(range(100)))), 100)

        self.assertEqual(MultiQueryModel.objects.filter(field1__in=list(range(100))).count(), 100)

        self.assertItemsEqual(
            MultiQueryModel.objects.filter(field1__in=list(range(100))).values_list("field1", flat=True),
            list(range(100)),
        )

        self.assertItemsEqual(
            MultiQueryModel.objects.filter(field1__in=list(range(100)))
            .order_by("-field1")
            .values_list("field1", flat=True),
            list(range(100))[::-1],
        )

    @override_settings(DJANGAE_MAX_QUERY_BRANCHES=10)
    def test_max_limit_enforced(self):
        for i in range(11):
            MultiQueryModel.objects.create(field1=i)

        self.assertRaises(NotSupportedError, list, MultiQueryModel.objects.filter(field1__in=list(range(11))))

    def test_pk_in_with_slicing(self):
        i1 = MultiQueryModel.objects.create()

        self.assertFalse(MultiQueryModel.objects.filter(pk__in=[i1.pk])[9999:])

        self.assertFalse(MultiQueryModel.objects.filter(pk__in=[i1.pk])[9999:10000])

    def test_limit_correctly_applied_per_branch(self):
        MultiQueryModel.objects.create(field2="test")
        MultiQueryModel.objects.create(field2="test2")

        with sleuth.watch("google.cloud.datastore.query.Query.fetch") as run_calls:

            list(MultiQueryModel.objects.filter(field2__in=["test", "test2"])[:1])

            self.assertEqual(1, run_calls.calls[0].kwargs["limit"])
            self.assertEqual(1, run_calls.calls[1].kwargs["limit"])

        with sleuth.watch("google.cloud.datastore.query.Query.fetch") as run_calls:

            list(MultiQueryModel.objects.filter(field2__in=["test", "test2"])[1:2])

            self.assertEqual(2, run_calls.calls[0].kwargs["limit"])
            self.assertEqual(2, run_calls.calls[1].kwargs["limit"])

    def test_ordered_by_nullable_field(self):
        NullableFieldModel.objects.create(pk=1)
        NullableFieldModel.objects.create(pk=5, nullable=2)

        results = NullableFieldModel.objects.filter(
            Q(nullable=1) | Q(nullable=2) | Q(nullable__isnull=True)
        ).order_by("nullable").values_list("pk", flat=True)

        self.assertCountEqual(results, [1, 5])



from . import TestCase
from .models import TestUser


class DeleteTestCase(TestCase):
    def test_entity_deleted(self):
        """Testing the basic `delete()` ORM interaction."""

        user_one = TestUser.objects.create(username="A", first_name="A", second_name="B")
        self.assertEqual(TestUser.objects.count(), 1)

        user_one.delete()

        with self.assertRaises(TestUser.DoesNotExist):
            user_one.refresh_from_db()

        with self.assertRaises(TestUser.DoesNotExist):
            TestUser.objects.get(username="A")

        self.assertEqual(TestUser.objects.count(), 0)

    def test_cache_keys_deleted(self):
        """FIXME-GCG"""
        pass

    def test_bulk_delete(self):
        """Testing the basic `delete()` ORM interaction."""
        TestUser.objects.create(username="One", first_name="A", second_name="B")
        TestUser.objects.create(username="Two", first_name="B", second_name="B")
        TestUser.objects.create(username="Three", first_name="C", second_name="B")

        self.assertEqual(TestUser.objects.count(), 3)

        TestUser.objects.all().delete()
        self.assertEqual(TestUser.objects.count(), 0)
